#!/usr/local/bin/wish8.3
#=======================================================================#
# SCRIPT  : satguitool.tcl                                              #
# PURPOSE : Small Application Toolkit User Interface Development Tool.  #
# AUTHOR  : Copyright (c) Adrian Davis 2000                             #
#-----------------------------------------------------------------------#
# HISTORY : Feb01 ALPHA3 - Third alpha release.                         #
#=======================================================================#

#=======================================================================#
# Name      : showError                                                 #
# Purpose   : Display "fatal" error messages on message line.           #
# Parameters: Message: Message text.                                    #
#=======================================================================#

proc showError {Message} {
   setMessage [list $Message red]
}

#=======================================================================#
# Name      : createConnection                                          #
# Purpose   : Create a socket connection to the SAT Application Host.   #
# Parameters: Host: Name of SAT Application Host.                       #
#           : Port: SAT Application Port.                               #
#=======================================================================#

proc createConnection {Host Port} {
   if [catch {set Connection [socket $Host $Port]}] {
      showError "Server not responding"
   }
   fconfigure $Connection -buffering line
   return $Connection
}

#=======================================================================#
# Name      : connectToServer                                           #
# Purpose   : Connect to the SAT Application Host.                      #
# Parameters: Host: Name of SAT Application Host.                       #
#           : Port: SAT Application Port.                               #
#=======================================================================#

proc connectToServer {Host Port} {
   set Server [createConnection $Host $Port]
   fileevent $Server readable [list readServer $Server]
   return $Server
}

#=======================================================================#
# Name      : readServer                                                #
# Purpose   : Read command from the Host and execute using safe parser. #
# Parameters: Server: Handle for server connection.                     #
#=======================================================================#

proc readServer {Server} {
   global PARSER

   if {[gets $Server Command] < 0} {
      catch {close $Server}
      showError "Lost server connection"
   } else {
      regsub -all "\034" $Command "\n" Command
      $PARSER eval $Command
   }
}

#=======================================================================#
# Name      : lock                                                      #
# Purpose   : Lock window for update/events.                            #
# Parameters: None.                                                     #
#=======================================================================#

proc lock {} {
   global CURSOR FOCUS LOCK

   set LOCK 1

   focus .busylock
   grab  .busylock
   set CURSOR [ . cget -cursor]
   . configure -cursor watch
   update
}

#=======================================================================#
# Name      : unlock                                                    #
# Purpose   : Unlock window for update/events.                          #
# Parameters: None.                                                     #
#=======================================================================#

proc unlock {} {
   global CURSOR FOCUS LOCK

   set LOCK 0

   . configure -cursor $CURSOR
   grab release .busylock
}

#=======================================================================#
# Name      : copySelection                                             #
# Purpose   : Copy fields from MultiList selection into data Items.     #
#=======================================================================#

proc copySelection {Name Copy} {
   global DATA

   if {[string match $Copy NONE] == 1} {return}
   if {! [info exists DATA($Name)]}    {return}

   set CopyItems [split $Copy ,]
   set DataItems [split $DATA($Name) \t]

   set CopyItem 0

   foreach DataItem $DataItems {
      if {[string match [lindex $CopyItems $CopyItem] ""] == 0} {
          setItem [list [lindex $CopyItems $CopyItem] $DataItem]
      }
      incr CopyItem
   }
}

#=======================================================================#
# Name      : formatDate                                                #
# Purpose   : Formats date to mm/dd/yy[yy] format for validation.       #
# Parameters: Date : Date to be formatted.                              #
#=======================================================================#

proc formatDate {Date} {
   global DATEFORMAT

   set SplitDate [split $Date $DATEFORMAT(separator)]
   set Day       [lindex $SplitDate $DATEFORMAT(day)]
   set Month     [lindex $SplitDate $DATEFORMAT(month)]
   set Year      [lindex $SplitDate $DATEFORMAT(year)]

   return "$Month/$Day/$Year"
}

#=======================================================================#
# Name      : formatOption                                              #
# Purpose   : Formats option to extract parameter(s).                   #
# Parameters: Option : Option to be formatted.                          #
#=======================================================================#

proc formatOption {Option} {

regexp {([^(]+)\(([^)]*)\)} $Option Match Dummy Parameter

return $Parameter

}

#=======================================================================#
# Name      : formatText                                                #
# Purpose   : Formats label text ("aaa_bbb_ccc" to "aaa bbb ccc").      #
# Parameters: Text : Text to be formatted.                              #
#=======================================================================#

proc formatText {Text} {
   regsub -all {^!.*}   $Text {}  Text
   regsub -all {^.*:}   $Text {}  Text
   regsub -all {[_]}    $Text { } Text
   return $Text
}

#=======================================================================#
# Name      : initGUI                                                   #
# Purpose   : Initialise application parameters.                        #
# Parameters: Title  : Window title.                                    #
#           : Menu   : Menubar options if required.                     #
#=======================================================================#

proc initGUI {Title} {
   global INITIALISING PLATFORM TITLE

   set TITLE         $Title
   set INITIALISING  0
}

#=======================================================================#
# Proc   : doAction                                                     #
# Purpose: Generate "action" command and send to server.                #
# Format : action Name Item                                             #
#=======================================================================#

proc doAction {Name} {
   global ACTION DATA GROUP LOCK MENUACTION SAFE SERVER VALIDATE

   if $LOCK {return}

   set ActionInfo ""

   if {[info exists ACTION($Name)]}     {set ActionInfo $ACTION($Name)}
   if {[info exists MENUACTION($Name)]} {set ActionInfo $MENUACTION($Name)}
   
   if {[string compare $ActionInfo ""] != 0} {

      if {$VALIDATE($Name,ACTION) && [info exists VALIDATE(LIST)]} {
         foreach Item $VALIDATE(LIST) {
            if {[info exists VALIDATE($Name,PREFIX)]} {
               set ItemName [lindex [split $Item .] 3]
               if {[string match $VALIDATE($Name,PREFIX) ""] == 1} {
                   if {[regexp {:} $ItemName] == 1} {continue}
               } else {
                  if {[string match $VALIDATE($Name,PREFIX):* $ItemName] == 0} {
                     continue
                  }
               }
            }
            if {[validateEntry $Item] == 0} {
               setFocus $Item.entry
               return 0
            }
         }
      }

      regsub -all "%" "$ActionInfo" $GROUP ActionInfo

      set Action    [lindex $ActionInfo 0]
      set Variables [lrange $ActionInfo 1 end]
 
      if {[string compare $Action EXIT]  == 0} {exit}

      foreach Variable $Variables {
         if {[info exists DATA($Variable)]} {
            set Action "$Action {$DATA($Variable)}"
         } elseif {[info exists SAFE($Variable)]} {
            set Action "$Action {$SAFE($Variable)}"
         }
      }

#     lock
      puts "ACTION=$Action"
#     puts $SERVER $Action
   } else {
      puts "Action does not exist"
   }
}

#=======================================================================#
# Name        : Unique name for widget.                                 #
# Options     : Optional parameters (See Options).                      #
# PositionX,Y : X,Y position of top left of (compound) widget.          #
# Size        : Format depends on widget type.                          #
# Text        : Text to appear on widget (label).                       #
# Title       : Title to appear on border/outline.                      #
#=======================================================================#

#=======================================================================#
# Proc   : action                                                       #
# Purpose: Pass-through action request to server.                       #
# Format : action {Action}                                              #
# Options: None.                                                        #
#=======================================================================#

proc action {Args} {
   global SERVER

   puts $SERVER "$Args"
}

#=======================================================================#
# Proc   : addBorder                                                    #
# Purpose: Add border with optional title.                              #
# Format : addBorder {Name PositionX,Y WidthxHeight Options}            #
# Options: groove  : Set relief to "groove".                            #
#        : outline : Set relief to default.                             #
#        : raised  : Set relief to "raised".                            #
#        : sunken  : Set relief to "sunken".                            #
#        : title   : Set title.                                         #
#=======================================================================#

proc addBorder {Args} {
   global ITEM MAIN RELIEF XFACTOR YFACTOR

   set Name      [lindex $Args 0]
   set Position  [lindex $Args 1]
   set Size      [lindex $Args 2]
   set Options   [string tolower [lrange $Args 3 end]]
   set PosX      [expr [lindex [split $Position ,] 0] * $XFACTOR]
   set PosY      [expr [lindex [split $Position ,] 1] * $YFACTOR]
   set Width     [expr [lindex [split $Size x] 0] * $XFACTOR]
   set Height    [expr [lindex [split $Size x] 1] * $YFACTOR]
   set NameText  [formatText $Name]
   set Name      [string tolower $Name]
   set Frame     $MAIN.$Name

   set ITEM($Name) 2

   set Relief $RELIEF

   foreach Option $Options {
      switch -- $Option {
         groove  {set Relief groove}
         outline {set Relief $RELIEF}
         raised  {set Relief raised}
         sunken  {set Relief sunken}
         title   {set Title $NameText}
      }
   }

   if {$Height == 0} {set Height 2}

   frame $Frame -borderwidth 2 -height $Height -relief $Relief -width $Width

   place $Frame -x $PosX -y $PosY

   if {[info exists Title]} {
      label $Frame-title -anchor center -text $Title
      place $Frame-title -x [expr {$PosX + 5}] -y [expr {$PosY - 11}]
   }
}

#=======================================================================#
# Proc   : addButton                                                    #
# Purpose: Add button.                                                  #
# Format : addButton {Name PositionX,Y Size Options}                    #
# Options: disabled: Set button state to "disabled".                    #
#        : exit    : Set button action to "EXIT".                       #
#        : noborder: Set button border to "0".                          #
#        : normal  : Set button state to "normal".                      #
#        : outline : Draw outline around button.                        #
#        : state   : Set button state to default.                       #
#        : DEFAULT : Set action.                                        #
#=======================================================================#

proc addButton {Args} {
   global ACTION ITEM MAIN RELIEF STATE VALIDATE XFACTOR YFACTOR

   set Name     [lindex $Args 0]
   set Position [lindex $Args 1]
   set Size     [lindex $Args 2]
   set Options  [string tolower [lrange $Args 3 end]]
   set PosX     [expr [lindex [split $Position ,] 0] * $XFACTOR]
   set PosY     [expr [lindex [split $Position ,] 1] * $YFACTOR]
   set State    $STATE

   if [regsub -- {^[+]} $Name "" Name] {set State normal}
   if [regsub -- {^[-]} $Name "" Name] {set State disabled}

   set NameText [formatText $Name]
   set Name     [string tolower $Name]
   set Frame    $MAIN.$Name

   set ITEM($Name) 0
   set VALIDATE($Name,ACTION) 0

   set Border 2
   set Relief flat

   foreach Option $Options {
      switch -glob -- $Option {
         disabled   {set State disabled}
         exit       {set ACTION($Name) EXIT}
         noborder   {set Border 0}
         normal     {set State normal}
         outline    {set Relief $RELIEF}
         state      {set State $STATE}
         validate   {set VALIDATE($Name,ACTION) 1}
         validate(* {set VALIDATE($Name,ACTION) 1; set VALIDATE($Name,PREFIX) [formatOption $Option]}
         default    {set ACTION($Name) $Option}
      }
   }

   frame $Frame -borderwidth $Border -relief $Relief

   set Bind1 "doAction $Name"

   button $Frame.button -command $Bind1 -state $State -text $NameText -width $Size

   pack  $Frame.button
   place $Frame -x $PosX -y $PosY
}

#=======================================================================#
# Proc   : addButtons                                                   #
# Purpose: Add multiple buttons in "row" layout.                        #
# Format : addButtons {PositionX,Y Size ?Pad? ButtonData}               #
# Options: DEFAULT : None.                                              #
#=======================================================================#

proc addButtons {Args} {
   set Position [lindex $Args 0]
   set Size     [lindex $Args 1]
   set Buttons  [lrange $Args 2 end]
   set PosX     [lindex [split $Position ,] 0]
   set PosY     [lindex [split $Position ,] 1]
   set ButtonX  $PosX
   set ButtonY  $PosY

   foreach Button $Buttons {
      if {[llength $Button] != 0} {
         set Name          [lindex $Button 0]
         set ButtonOptions [lrange $Button 1 end]
         if {[regexp {^[0-9]+$} $Name]} {
            set ButtonX [expr {$ButtonX + $Name}]
         } else {
            addButton "[list $Name $ButtonX,$ButtonY $Size] $ButtonOptions"
            set ButtonX [expr {$ButtonX + $Size + 4}]
         }
      } else {
         set ButtonX [expr {$ButtonX + $Size}]
      }
   }
}

#=======================================================================#
# Proc   : addCheckButtons                                              #
# Purpose: Add "check buttons".                                         #
# Format : addCheckButtons {Name PositionX,Y {Labels} Options}          #
#        : Labels    : [+]Label                                         #
#        : +         : Set button with this label to be selected.       #
# Options: disabled  : Disable check buttons.                           #
#        : horizontal: Line-up buttons horizontally (Default vertical). #
#        : normal    : Enable check buttons.                            #
#        : outline   : Draw outline around check buttons.               #
#        : title     : Set title.                                       #
#=======================================================================#

proc addCheckButtons {Args} {
   global DATA ITEM MAIN RELIEF STATE XFACTOR YFACTOR

   set Name     [lindex $Args 0]
   set Position [lindex $Args 1]
   set Labels   [lindex $Args 2]
   set Options  [string tolower [lrange $Args 3 end]]
   set PosX     [expr [lindex [split $Position ,] 0] * $XFACTOR]
   set PosY     [expr [lindex [split $Position ,] 1] * $YFACTOR]
   set State    $STATE

   if [regsub -- {^[+]} $Name "" Name] {set State normal}
   if [regsub -- {^[-]} $Name "" Name] {set State disabled}

   set NameText [formatText $Name]
   set Name     [string tolower $Name]
   set Frame    $MAIN.$Name

   set ITEM($Name) 1

   set Relief flat
   set Side   top

   foreach Option $Options {
      switch -- $Option {
         disabled   {set State disabled}
         horizontal {set Side left}
         normal     {set State normal}
         outline    {set Relief $RELIEF}
         title      {set Title $NameText}
      }
   }

   frame $Frame -borderwidth 2 -relief $Relief

   foreach Label $Labels {
      set Active 0
      if {[string compare [string range $Label 0 0] "+"] == 0} {
         set Label [string range $Label 1 end]
         set Active 1
      }

      set Button              [string tolower $Label]
      set LabelText           [formatText $Label]
      set ITEM($Name:$Button) 1

      checkbutton $Frame.$Button -offvalue N -onvalue Y -state $State -text $LabelText -variable DATA($Name:$Button)
      $Frame.$Button configure -activebackground [$Frame.$Button cget -background]
      $Frame.$Button configure -highlightthickness 0
      pack $Frame.$Button -anchor w -padx 2 -pady 4 -side $Side
      if $Active {$Frame.$Button select}
   }

   place $Frame -x $PosX -y $PosY

   if {[info exists Title]} {
      label $Frame-title -anchor center -text $Title
      place $Frame-title -x [expr $PosX + 5] -y [expr $PosY - 13]
   }
}

#=======================================================================#
# Proc   : addData                                                      #
# Purpose: Adds DATA -or- SAFE data item.                               #
# Format : addData {Name Value Options}                                 #
# Options: safe : Set as SAFE data (Will not be removed by "clear").    #
#=======================================================================#

proc addData {Args} {
   global DATA ITEM SAFE

   set Name    [string tolower [lindex $Args 0]]
   set Value   [lindex $Args 1]
   set Options [string tolower [lrange $Args 2 end]]

   set ITEM($Name) 5

   set Safe 0

   foreach Option $Options {
      switch -- $Option {
         safe {set Safe 1}
      }
   }

   if $Safe {
      set SAFE($Name) $Value
   } else {
      set DATA($Name) $Value
   }
}

#=======================================================================#
# Proc   : addEntry                                                     #
# Purpose: Add entry box with label.                                    #
# Format : addEntry {Name PositionX,Y Size Options}                     #
# Options: center  : Centre justify label.                              #
#        : disabled: Set entry state to "disabled".                     #
#        : focus   : Set focus to this entry.                           #
#        : normal  : Set entry state to "normal".                       #
#        : outline : Draw outline around entry.                         #
#        : right   : Right justify label.                               #
#        : secret  : Display "*" for each entry character.              #
#        : DEFAULT : Set action/validation.                             #
#=======================================================================#

proc addEntry {Args} {
   global ACTION DATA ITEM MAIN RELIEF STATE VALIDATE VALIDATION XFACTOR YFACTOR

   set Name      [lindex $Args 0]
   set Position  [lindex $Args 1]
   set Size      [lindex $Args 2]
   set Options   [string tolower [lrange $Args 3 end]]
   set PosX      [expr [lindex [split $Position ,] 0] * $XFACTOR]
   set PosY      [expr [lindex [split $Position ,] 1] * $YFACTOR]
   set SizeLabel [lindex [split $Size ,] 0]
   set SizeEntry [lindex [split $Size ,] 1]
   set State     $STATE

   if [regsub -- {^[+]} $Name "" Name] {set State normal}
   if [regsub -- {^[-]} $Name "" Name] {set State disabled}

   set NameText  [formatText $Name]
   set Name      [string tolower $Name]

   regsub -all {^!} $Name {} Name

   set Frame $MAIN.$Name

   set ITEM($Name) 3
   set VALIDATE($Name,ACTION) 0
   set VALIDATE($Name,TEXT) $NameText

   set Anchor   w
   set Focus    0
   set Relief   flat
   set Secret   0
   set Validate none

   if {[string compare $State normal] == 0} {
      set Background white
   } else {
      set Background lightgrey
   }

   foreach Option $Options {
      switch -- $Option {
         center     {set Anchor center}
         disabled   {set State disabled; set Background lightgrey}
         focus      {set Focus 1}
         normal     {set State normal; set Background white}
         outline    {set Relief $RELIEF}
         right      {set Anchor e}
         secret     {set Secret 1}
         validate   {set VALIDATE($Name,ACTION) 1}
         validate(* {set VALIDATE($Name,ACTION) 1; set VALIDATE($Name,PREFIX) [formatOption $Option]}
         default    {
            if {[info exists VALIDATION($Option)]} {
               lappend VALIDATE(LIST) $Frame
               set VALIDATE($Name) $VALIDATION($Option)
               set Validate focusout
            } else {
               set ACTION($Name) $Option
               set Bind1 "doAction $Name"
            }
         }
      }
   }

   frame $Frame -borderwidth 2 -relief $Relief
   label $Frame.label -anchor $Anchor -text $NameText -width $SizeLabel
   entry $Frame.entry -background $Background -invcmd "setFocus %W" -state $State -textvariable DATA($Name) -validate $Validate -vcmd "validateEntry %W" -width $SizeEntry

   if {[info exists ACTION($Name)]} {
      bind $Frame.entry <Return> "doAction $Name"
   }

   pack  $Frame.label -side left
   pack  $Frame.entry -side left
   place $Frame -x $PosX -y $PosY

   if $Secret {$Frame.entry configure -show "*"}
   if $Focus  {focus $Frame.entry}
}


#=======================================================================#
# Proc   : addEntries                                                   #
# Purpose: Add multiple entry boxes in "grid" layout.                   #
# Format : addEntries {PositionX,Y ?Pad? ColumnData}                    #
# Options: DEFAULT : None.                                              #
#=======================================================================#

proc addEntries {Args} {
   set Position [lindex $Args 0]

   if {[regexp {^[0-9]*$} [lindex $Args 1]] == 1} {
      set Pad     [lindex $Args 1]
      set Columns [lrange $Args 2 end]
   } else {
      set Pad 5
      set Columns [lrange $Args 1 end]
   }

   set PosX   [lindex [split $Position ,] 0]
   set PosY   [lindex [split $Position ,] 1]
   set EntryX $PosX

   foreach Column $Columns {
      set Size         [lindex $Column 0]
      set Entries      [lrange $Column 1 end]
      set SizeLabel    [lindex [split $Size ,] 0]
      set MaxSizeEntry [lindex [split $Size ,] 1]
      set EntryY       $PosY
      foreach Entry $Entries {
         set Name [lindex $Entry 0]
         if {[llength $Entry] !=0} {
            if {[llength $Entry] > 1} {
               if {[regexp {^[0-9]*$} [lindex $Entry 1]] == 1} {
                  set SizeEntry [lindex $Entry 1]
                  set EntryOptions [lrange $Entry 2 end]
               } else {
                  set SizeEntry $MaxSizeEntry
                  set EntryOptions [lrange $Entry 1 end]
               }
            } else {
               set SizeEntry $MaxSizeEntry
               set EntryOptions ""
            }
            addEntry "[list $Name $EntryX,$EntryY $SizeLabel,$SizeEntry] $EntryOptions"
         }  
         set EntryY [expr {$EntryY + 2}]
      }
      set EntryX [expr {$EntryX + $SizeLabel + $MaxSizeEntry + $Pad}]
   }
}

#=======================================================================#
# Proc   : addLabel                                                     #
# Purpose: Add label (text).                                            #
# Format : addLabel {Name PositionX,Y Size Text Options}                #
# Options: blue    : Set text colour to "blue".                         #
#        : center  : Centre justify label.                              #
#        : outline : Draw outline around entry.                         #
#        : noborder: Set label border to "0".                           #
#        : red     : Set text colour to "red".                          #
#        : right   : Right justify label.                               #
#=======================================================================#

proc addLabel {Args} {
   global ITEM MAIN RELIEF XFACTOR YFACTOR

   set Name      [lindex $Args 0]
   set Position  [lindex $Args 1]
   set SizeLabel [lindex $Args 2]
   set LabelText [lindex $Args 3]
   set Options   [string tolower [lrange $Args 4 end]]
   set PosX      [expr [lindex [split $Position ,] 0] * $XFACTOR]
   set PosY      [expr [lindex [split $Position ,] 1] * $YFACTOR]
   set NameText  [formatText $Name]
   set Name      [string tolower $Name]
   set Frame $MAIN.$Name

   set ITEM($Name) 4

   set Anchor     w
   set Border     2
   set Foreground black
   set Justify    left
   set Relief     flat

   foreach Option $Options {
      switch -- $Option {
         blue     {set Foreground blue}
         center   {set Anchor center;set Justify center}
         outline  {set Relief $RELIEF}
         noborder {set Border 0}
         red      {set Foreground red}
         right    {set Anchor e;set Justify right}
      }
   }

   frame $Frame -borderwidth $Border -relief $Relief
   label $Frame.label -anchor $Anchor -foreground $Foreground -justify $Justify -text $LabelText -width $SizeLabel
   pack  $Frame.label -side left
   place $Frame -x $PosX -y $PosY
}

#=======================================================================#
# Proc   : addMenu                                                      #
# Purpose: Add menu bar.                                                #
# Format : addMenu {MenuData}                                           #
#=======================================================================#

proc addMenu {Args} {
   global GROUP HEIGHT MENU MENUACTION MENULIST PLATFORM STATE VALIDATE WIDTH

   if {[string compare $Args ""] == 0} {return}
   if {[info exists MENU(.menubar:1)]} {return}

   update

   . configure -menu .menubar
   menu .menubar

   set MenuIndex 1

   foreach Menu $Args {
      set Name      [lindex $Menu 0]
      set MenuItems [lrange $Menu 1 end]

      set MenuState     $STATE
      set StateOverride 0
      if [regsub -- {^[+]} $Name "" Name] {set MenuState normal;set StateOverride 1}
      if [regsub -- {^[-]} $Name "" Name] {set MenuState disabled; set StateOverride 1}

      set Text      [formatText $Name]
      set Name      [string tolower $Name]

      if $StateOverride {
         set MENU(.menubar:$MenuIndex) $MenuState 
      } else {
         set MENU(.menubar:$MenuIndex) state
      }

      incr MenuIndex

      .menubar add cascade -label $Text -menu .menubar.$Name -state $MenuState
       menu .menubar.$Name
      .menubar.$Name configure -tearoff 0

      set  ItemIndex 0

      foreach MenuItem $MenuItems {
         set Item    [lindex $MenuItem 0]
         set Options [lrange $MenuItem 1 end]

         set ItemState     $STATE
         set StateOverride 0
         if [regsub -- {^[+]} $Item "" Item] {set ItemState normal; set StateOverride 1}
         if [regsub -- {^[-]} $Item "" Item] {set ItemState disabled; set StateOverride 1}

         set Label   [formatText $Item]
         set Item    [string tolower $Item]

         if {[string compare $Item ""] == 0} {
            .menubar.$Name add separator
         } else  {
            set VALIDATE($Name:$Item,ACTION) 0
            foreach Option $Options {
               switch -glob -- $Option {
                  disabled   {set ItemState disabled; set StateOverride 1}
                  exit       {set MENUACTION($Name:$Item) EXIT}
                  group(*    {set ItemState [formatOption $Option];set StateOverride 1}
                  normal     {set ItemState normal; set StateOverride 1}
                  validate   {set VALIDATE($Name:$Item,ACTION) 1}
                  validate(* {set VALIDATE($Name:$Item,ACTION) 1; set VALIDATE($Name:$Item,PREFIX) [formatOption $Option]}
                  default    {set MENUACTION($Name:$Item) $Option}
               }
            }

            lappend MENULIST $Name:$Item

            if $StateOverride {
               set MENU(.menubar.$Name:$ItemIndex) $ItemState 
            } else {
               set MENU(.menubar.$Name:$ItemIndex) state
            }

            set Command "doAction $Name:$Item"
            .menubar.$Name add command -command "$Command" -label "$Label"
         }
         incr ItemIndex
      }
   }
   .message configure -height 20
   .line configure -relief groove
   refreshMenu
}

#=======================================================================#
# Proc   : addMultiList                                                 #
# Purpose: Add multi column list box.                                   #
# Format : addMultiList {Name PositionX,Y Size {Columns} Options}       #
#        : Columns : Name,Size[,Type] ...                               #
#        : Type    : right   : Right justify.                           #
#        :         : numeric : Numeric justify.                         #
# Options: outline : Draw outline around entry.                         #
#        : yscroll : Add "Y" scrollbar.                                 #
#        : DEFAULT : Set action.                                        #
#=======================================================================#

proc addMultiList {Args} {
   global ACTION DATA FONT ITEM MAIN RELIEF VALIDATE XFACTOR YFACTOR

   set Name      [lindex $Args 0]
   set Position  [lindex $Args 1]
   set SizeText  [lindex $Args 2]
   set Columns   [lindex $Args 3]
   set Options   [string tolower [lrange $Args 4 end]]
   set PosX      [expr [lindex [split $Position ,] 0] * $XFACTOR]
   set PosY      [expr [lindex [split $Position ,] 1] * $YFACTOR]
   set NameText  [formatText $Name]
   set Name      [string tolower $Name]

   set ITEM($Name) 6
   set VALIDATE($Name,ACTION) 0

   set DATA(List,$Name) ""

   set Copy      NONE
   set End       1
   set NoYScroll 1
   set Number    0
   set Relief    flat
   set Tabs      ""
   set Width     1

   foreach Option $Options {
      switch -glob -- $Option {
         copy(*     {set Copy [formatOption $Option]}
         outline    {set Relief $RELIEF}
         validate   {set VALIDATE($Name,ACTION) 1}
         validate(* {set VALIDATE($Name,ACTION) 1; set VALIDATE($Name,PREFIX) [formatOption $Option]}
         yscroll    {set NoYScroll 0}
         default    {set ACTION($Name) $Option}
      }
   }

   set Frame      [frame $MAIN.$Name -borderwidth 2 -relief $Relief]
   set LabelFrame [frame $Frame.labels]

   foreach Column $Columns {
      set Anchor     left
      set ColumnList [split $Column ,]
      set LabelText  [lindex $ColumnList 0]
      set SizeColumn [lindex $ColumnList 1]

      if {[llength $ColumnList] == 3} {
         set Anchor [lindex $ColumnList 2]
      }

      set Width [expr {$Width + ($SizeColumn * $XFACTOR) + 6}]
      set End   [expr {$End + ($SizeColumn * $XFACTOR) + 6}]

      switch -- $Anchor {
         left    {set Tab [expr {$Width - (($SizeColumn * $XFACTOR) + 6)}]}
         numeric {set Tab [expr {$Width - ($XFACTOR * 4)}]}
         right   {set Tab [expr {$Width - ($XFACTOR + 3)}]}
         default {set Tab [expr {$Width - (($SizeColumn * $XFACTOR) + 6)}]}
      }

      set Tabs       "$Tabs $Tab $Anchor"
      set Label      [string tolower $LabelText]
      set LabelText  [formatText $LabelText]    

      frame $LabelFrame.$Label -borderwidth 2 -height 20 -relief raised -width [expr {($SizeColumn * $XFACTOR) + 6}]
      label $LabelFrame.$Label.text -text $LabelText
      pack propagate $LabelFrame.$Label false
      pack $LabelFrame.$Label -side left
      pack $LabelFrame.$Label.text -side left

      incr Number
   }

   set LastLabelWidth [$LabelFrame.$Label cget -width]
   set Filler         [expr {$XFACTOR - ($Number % $XFACTOR)}]
   set LastLabelWidth [expr {$LastLabelWidth + $Filler}]
   $LabelFrame.$Label configure -width $LastLabelWidth

   set End   [expr {$End - $XFACTOR}]
   set Tabs  "$Tabs $End left"
   set Width [expr {$Width / $XFACTOR}]

   text $Frame.text -background white -cursor arrow -font $FONT -height $SizeText -tabs $Tabs -width $Width -wrap none -yscrollcommand [list $Frame.ybar set]
   scrollbar $Frame.ybar -command [list $Frame.text yview]

   grid $LabelFrame -row 0 -column 0 -sticky w
   grid $Frame.text -row 1 -column 0 -sticky e
   grid $Frame.ybar -row 0 -column 1 -rowspan 2 -sticky nsew

   if $NoYScroll {grid forget $Frame.ybar}

   $Frame.text configure -state disabled
   $Frame.text tag configure selection -background blue -foreground white

   set Select "$Frame.text tag remove selection 1.0 end;$Frame.text tag add selection \"@%x,%y linestart\" \"@%x,%y lineend\""
   set Set    "catch {set DATA($Name) \[$Frame.text get \"selection.first +1 char\" \"selection.last -1 char\"\]}"
   set Bind1  "$Select;$Set;copySelection $Name $Copy"
   set Bind2  "doAction $Name"

   bind $Frame.text <Button-1> $Bind1
   bind $Frame.text <Double-1> $Bind2

   place $Frame -x $PosX -y $PosY
}

#=======================================================================#
# Proc   : addRadioButtons                                              #
# Purpose: Add "radio buttons".                                         #
# Format : addRadioButtons {Name PositionX,Y {Labels} Options}          #
#        : Labels    : [+]Label                                         #
#        : +         : Set button with this label to be selected.       #
# Options: disabled  : Disable radio buttons.                           #
#        : horizontal: Line-up buttons horizontally (Default vertical). #
#        : normal    : Enable radio buttons.                            #
#        : outline   : Draw outline around radio buttons.               #
#        : title     : Set title.                                       #
#=======================================================================#

proc addRadioButtons {Args} {
   global DATA ITEM MAIN RELIEF STATE XFACTOR YFACTOR

   set Name     [lindex $Args 0]
   set Position [lindex $Args 1]
   set Labels   [lindex $Args 2]
   set Options  [string tolower [lrange $Args 3 end]]
   set PosX     [expr [lindex [split $Position ,] 0] * $XFACTOR]
   set PosY     [expr [lindex [split $Position ,] 1] * $YFACTOR]
   set State    $STATE

   if [regsub -- {^[+]} $Name "" Name] {set State normal}
   if [regsub -- {^[-]} $Name "" Name] {set State disabled}

   set NameText [formatText $Name]
   set Name     [string tolower $Name]
   set Frame    $MAIN.$Name

   set ITEM($Name) 7

   set Relief flat
   set Side   top

   foreach Option $Options {
      switch -- $Option {
         disabled   {set State disabled}
         horizontal {set Side left}
         normal     {set State normal}
         outline    {set Relief $RELIEF}
         title      {set Title $NameText}
      }
   }

   frame $Frame -borderwidth 2 -relief $Relief

   foreach Label $Labels {
      set Active 0
      if {[string compare [string range $Label 0 0] "+"] == 0} {
         set Label [string range $Label 1 end]
         set Active 1
      }

      set Button     [string tolower $Label]
      set LabelText  [formatText $Label]

      radiobutton $Frame.$Button -state $State -text $LabelText -value $Button -variable DATA($Name)
      $Frame.$Button configure -activebackground [$Frame.$Button cget -background]
      $Frame.$Button configure -highlightthickness 0
      pack $Frame.$Button -anchor w -padx 2 -pady 4 -side $Side
      if $Active {$Frame.$Button select}
   }

   place $Frame -x $PosX -y $PosY

   if {[info exists Title]} {
      label $Frame-title -anchor center -text $Title
      place $Frame-title -x [expr $PosX + 5] -y [expr $PosY - 13]
   }
}

#=======================================================================#
# Proc   : addValidation                                                #
# Purpose: Adds validation pattern.                                     #
# Format : addValidation {Name Pattern}                                 #
#        : Pattern : Regular expression pattern to use for validation.  #
#=======================================================================#

proc addValidation {Args} {
   global VALIDATION

   set Name    [string tolower [lindex $Args 0]]
   set Pattern [lindex $Args 1]

   set VALIDATION($Name) $Pattern
}

#=======================================================================#
# Proc   : clear                                                        #
# Purpose: Delete all widgets from "MAIN" and unset DATA array.         #
# Format : clear                                                        #
#=======================================================================#

proc clear {} {
   global ACTION DATA HEIGHT ITEM MAIN MENULIST VALIDATE WIDTH

   destroy $MAIN
   if {[info exists ACTION]}   {unset ACTION}
   if {[info exists DATA  ]}   {unset DATA}
   if {[info exists ITEM  ]}   {unset ITEM}

   if {[info exists VALIDATE]} {
      if {[info exists MENULIST]} {
         foreach ValidateItem [array names VALIDATE] {
            set NoMatch 1
            foreach MenuItem $MENULIST {
               set Pattern $MenuItem,
               if {[regexp $Pattern $ValidateItem]} {
                  set NoMatch 0
                  continue
               }
            }
            if $NoMatch {unset VALIDATE($ValidateItem)}
         }
      } else {
         unset VALIDATE
      }
   }

   frame $MAIN -width $WIDTH -height $HEIGHT

   pack  propagate $MAIN false
   pack  $MAIN -side top
}

#=======================================================================#
# Proc   : newScreen                                                    #
# Purpose: Initialises new screen.                                      #
# Format : newScreen {Title Options}                                    #
# Options: replace : Replace (rather than append) to window title.      #
#        : INTxINT : Set size of application window.                    #
#        : DEFAULT : Set Group.                                         #
#=======================================================================#

proc newScreen {Args} {

   set TitleArgs [lindex $Args 0]
   set Options   [lrange $Args 1 end]

   set Group ""
   set Size  ""

   foreach Option $Options {
      switch -regexp -- $Option {
         ^replace$       {lappend $TitleArgs replace}
         ^[0-9]*x[0-9]*$ {set Size  $Option}
         default         {set Group $Option}
      }
   }

   clear
   setTitle $TitleArgs
   setMessage {}

   if {[string compare $Size  ""] != 0} {setSize  $Size}
   if {[string compare $Group ""] != 0} {setGroup $Group}
}

#=======================================================================#
# Proc   : refreshMenu                                                  #
# Purpose: Refresh menu to use current State and Group.                 #
# Format : refreshMenu                                                  #
#=======================================================================#

proc refreshMenu {} {
   global GROUP MENU STATE

   foreach Item [array names MENU] {
      set Name      [lindex [split $Item :] 0]
      set Index     [lindex [split $Item :] 1]
      set ItemState $MENU($Item)

      switch -- $ItemState {
         disabled   {set ItemState disabled}
         normal     {set ItemState normal}
         state      {set ItemState $STATE}
         default    {
            foreach ItemGroup [split $ItemState ,] {
               set ItemState     disabled
               set StateOverride 0

               if [regsub -- {^[+]} $ItemGroup "" ItemGroup] {set OverrideState normal; set StateOverride 1}
               if [regsub -- {^[-]} $ItemGroup "" ItemGroup] {set OverrideState disabled; set StateOverride 1}

               if $StateOverride {
                  if {[string compare $ItemGroup $GROUP] == 0} {
                     set ItemState $OverrideState
                     break
                  }
               } else {
                  if {[string compare $ItemGroup $GROUP] == 0 && [string compare $STATE normal] == 0} {
                     set ItemState normal
                     break
                  }
               }
            }
         }
      }
      $Name entryconfigure $Index -state $ItemState
   }
}

#=======================================================================#
# Proc   : remove                                                       #
# Purpose: Delete widget and associated "DATA" -or- "SAFE" data item.   #
# Format : remove Name                                                  #
#=======================================================================#

proc remove {Args} {
   global ACTION DATA ITEM MAIN MENU MENUACTION MENULIST SAFE VALIDATE

   if {[string compare $Args menu] == 0} {
      . configure -menu {}
      destroy .menubar
      .line configure -relief flat
      if {[info exists VALIDATE] && [info exists MENULIST]} {
         foreach ValidateItem [array names VALIDATE] {
            foreach MenuItem $MENULIST {
               set Pattern $MenuItem,
               if {[regexp $Pattern $ValidateItem]} {unset VALIDATE($ValidateItem)}
            }
         }
      }
      unset MENU
      unset MENUACTION
      unset MENULIST
      return
   }

   set NamePattern [string tolower $Args]
   set Names       [array names ITEM $NamePattern]

   foreach Name $Names {
      set Frame $MAIN.$Name

      if {[info exists SAFE($Name)]} {
         unset SAFE($Name)
      } else {
         destroy $Frame
         destroy $Frame-title

         if {[info exists VALIDATE(LIST)]} {
            set Item [lsearch -exact $VALIDATE(LIST) $Frame]
            if {$Item >= 0} {set VALIDATE(LIST) [lreplace $VALIDATE(LIST) $Item $Item]}
         }

         set DataNames "$Name [array names DATA $Name:*]"
         foreach Name $DataNames {
            if {[info exists ACTION($Name)]}          {unset ACTION($Name)}
            if {[info exists DATA($Name)]}            {unset DATA($Name)}
            if {[info exists ITEM($Name)]}            {unset ITEM($Name)}
            if {[info exists VALIDATE($Name)]}        {unset VALIDATE($Name)}
            if {[info exists VALIDATE($Name,ACTION)]} {unset VALIDATE($Name,ACTION)}
            if {[info exists VALIDATE($Name,PREFIX)]} {unset VALIDATE($Name,PREFIX)}
            if {[info exists VALIDATE($Name,TEXT)]}   {unset VALIDATE($Name,TEXT)}
         }
      }
   }
}

#=======================================================================#
# Proc   : setAction                                                    #
# Purpose: Sets action for widgets which support actions.               #
# Format : setAction {Name Action}                                      #
#=======================================================================#

proc setAction {Args} {
   global ACTION

   set Name          [string tolower [lindex $Args 0]]
   set ACTION($Name) [lindex $Args 1]
}

#=======================================================================#
# Proc   : setButton                                                    #
# Purpose: Sets button state (normal/disabled).                         #
# Format : setButton {Name State}                                       #
#=======================================================================#

proc setButton {Args} {
   global MAIN

   set Name  [string tolower [lindex $Args 0]]
   set State [lindex $Args 1]
   set Frame $MAIN.$Name

   $Frame.button configure -state $State
}

#=======================================================================#
# Proc   : setCheckButtons                                              #
# Purpose: Sets check button selection.                                 #
# Format : setCheckButtons {Name Buttons}                               #
#        : setCheckButtons {Name:Button Value}                          #
#        : Buttons : [-]Button                                          #
#        : -       : Set button with this label to be deselected.       #
#        : Value   : Y/N                                                #
#=======================================================================#

proc setCheckButtons {Args} {
   global MAIN

   set Name    [string tolower [lindex $Args 0]]
   set Buttons [string tolower [lrange $Args 1 end]]

   if {[regexp {(.+):(.+)} $Name Match Name Button]} {
      if {[string compare [lindex $Buttons 0] "y"] == 0} {
         set Buttons $Button
      } else {
         set Buttons "-$Button"
      }
   }

   set Frame $MAIN.$Name

   foreach Button $Buttons {
      if {[string compare [string range $Button 0 0] "-"] == 0} {
         set Button [string range $Button 1 end]
         $Frame.$Button deselect
      } else {
         $Frame.$Button select
      }
   }
}

#=======================================================================#
# Proc   : setData                                                      #
# Purpose: Sets DATA -or- SAFE data item.                               #
# Format : setData {Name Value}                                         #
#=======================================================================#

proc setData {Args} {
   global DATA SAFE

   set Name    [string tolower [lindex $Args 0]]
   set Value   [lindex $Args 1]

   if {[info exists SAFE($Name)]} {
      set SAFE($Name) $Value
   } else {
      set DATA($Name) $Value
   }
}

#=======================================================================#
# Proc   : setDateFormat                                                #
# Purpose: Sets date format used for date validation.                   #
# Format : setDateFormat {Format}                                       #
#=======================================================================#

proc setDateFormat {Args} {
   global DATEFORMAT

   set DateFormat [string tolower $Args]

   if {[regsub -all -- {/} $DateFormat {} DateFormat] != 0} {
      set DATEFORMAT(separator) "/"
   } elseif {[regsub -all -- {[.]} $DateFormat {} DateFormat] != 0} {
      set DATEFORMAT(separator) "."
   } elseif {[regsub -all -- {-} $DateFormat {} DateFormat] != 0} {
      set DATEFORMAT(separator) "-"
   }

   switch -- $DateFormat {
      mmddyyyy {set DATEFORMAT(day) 1;set DATEFORMAT(month) 0;set DATEFORMAT(year) 2}
      ddmmyyyy {set DATEFORMAT(day) 0;set DATEFORMAT(month) 1;set DATEFORMAT(year) 2}
      yyyymmdd {set DATEFORMAT(day) 2;set DATEFORMAT(month) 1;set DATEFORMAT(year) 0}
      mmddyy   {set DATEFORMAT(day) 1;set DATEFORMAT(month) 0;set DATEFORMAT(year) 2}
      ddmmyy   {set DATEFORMAT(day) 0;set DATEFORMAT(month) 1;set DATEFORMAT(year) 2}
      yymmdd   {set DATEFORMAT(day) 2;set DATEFORMAT(month) 1;set DATEFORMAT(year) 0}
   }
}

#=======================================================================#
# Proc   : setEntry                                                     #
# Purpose: Sets value of entry.                                         #
# Format : setEntry {Name Value}                                        #
#=======================================================================#

proc setEntry {Args} {
   global DATA VALIDATE

   set Name  [string tolower [lindex $Args 0]]
   set Value [lindex $Args 1]
   set DATA($Name) $Value

   if {[info exists VALIDATE($Name)]} {
      after idle ".main.main.$Name.entry configure -validate focusout"
   }
}

#=======================================================================#
# Proc   : setFocus                                                     #
# Purpose: Sets focus to failed validation entry.                       #
# Format : setFocus item                                                #
#=======================================================================#

proc setFocus {Item} {
  set Focus [focus]

  if {[string compare {} $Focus] && \
      [string equal [winfo class $Focus] Entry] && \
      [regexp {^(focus(out)?|all)} [set Validate [$Focus cget -validate]]]} {
     $Focus configure -validate none
     after idle [list $Focus configure -validate $Validate]
  }
  focus $Item
}

#=======================================================================#
# Proc   : setGroup                                                     #
# Purpose: Sets group.                                                  #
# Format : setGroup Group                                               #
#=======================================================================#

proc setGroup {Args} {
   global GROUP

   set GROUP $Args

   refreshMenu
}

#=======================================================================#
# Proc   : setItem                                                      #
# Purpose: Sets value of "item" using defaults.                         #
# Format : setItem {Name Value}                                         #
#=======================================================================#

proc setItem {Args} {
   global ITEM

   set Name  [string tolower [lindex $Args 0]]
   set Value [lrange $Args 1 end]

   if {[info exists ITEM($Name)]} {
      set Item $ITEM($Name)
   } else {
      if {[string compare [string range $Value 0 0] "-"] == 0} {
         set Item $ITEM($Name:[string range $Value 1 end])
      } else {
         set Item $ITEM($Name:$Value)
      }
   }

   switch -- $Item {
      0 {setButton "$Name $Value"}
      1 {setCheckButtons "$Name $Value"}
      2 {puts "USED: Border"}
      3 {setEntry "$Name $Value"}
      4 {setLabel "$Name $Value"}
      5 {setData  "$Name $Value"}
      6 {setMultiList "$Name $Value"}
      7 {setRadioButtons "$Name $Value"}
      8 {puts "Text"}
   }
}

#=======================================================================#
# Proc   : setItems                                                     #
# Purpose: Sets values for list of widgets.                             #
# Format : setItems {{Name Value} {Name Value} ...}                     #
#=======================================================================#

proc setItems {Args} {
   global DATA

   foreach Set $Args {
      set Name  [lindex $Set 0]
      set Value [lindex $Set 1]
      setItem [list $Name $Value]
   }
}

#=======================================================================#
# Proc   : setLabel                                                     #
# Purpose: Sets label text.                                             #
# Format : setLabel {Name Label}                                        #
#=======================================================================#

proc setLabel {Args} {
   global MAIN

   set Name  [string tolower [lindex $Args 0]]
   set Label [lindex $Args 1]
   set Frame $MAIN.$Name

   $Frame.label configure -text $Label
}

#=======================================================================#
# Proc   : setMessage                                                   #
# Purpose: Sets message line text.                                      #
# Format : setMessage {Message Options}                                 #
# Options: blue : Set text colour to "blue".                            #
#        : red  : Set text colour to "red".                             #
#=======================================================================#

proc setMessage {Args} {
   set Text    [lindex $Args 0]
   set Options [string tolower [lrange $Args 1 end]]

   set Colour black

   foreach Option $Options {
      switch -- $Option {
         blue {set Colour blue}
         red  {set Colour red}
      }
   }

   .message.text configure -text $Text -foreground $Colour
}

#=======================================================================#
# Proc   : setMultiList                                                 #
# Purpose: Sets multi column listbox data from list.                    #
# Format : setMultiList {name {tab demilited line} ...}                 #
#=======================================================================#

proc setMultiList {Args} {
   global MAIN

   set Name  [string tolower [lindex $Args 0]]
   set List  [lrange $Args 1 end]
   set Frame $MAIN.$Name

   $Frame.text configure -state normal
   $Frame.text delete 1.0 end

   foreach Item $List {
      $Frame.text insert end "\t$Item\t\n"
   }

   $Frame.text delete "end -1 char" end
   $Frame.text configure -state disabled
}

#=======================================================================#
# Proc   : setRadioButtons                                              #
# Purpose: Sets radio button selection.                                 #
# Format : setRadioButtons {Name Button}                                #
#=======================================================================#

proc setRadioButtons {Args} {
   global MAIN

   set Name   [string tolower [lindex $Args 0]]
   set Button [string tolower [lindex $Args 1]]
   set Frame  $MAIN.$Name

   $Frame.$Button select
}

#=======================================================================#
# Proc   : setSize                                                      #
# Purpose: Sets screen size.                                            #
# Format : setSize WidthxHeight                                         #
#=======================================================================#

proc setSize {Args} {
   global HEIGHT MAIN XFACTOR WIDTH YFACTOR

   set WIDTH  [expr [lindex [split $Args x] 0] * $XFACTOR]
   set HEIGHT [expr [lindex [split $Args x] 1] * $YFACTOR]

   $MAIN configure -height $HEIGHT -width $WIDTH
   .main configure -height $HEIGHT -width $WIDTH
   .line configure -width $WIDTH

   update
}

#=======================================================================#
# Proc   : setState                                                     #
# Purpose: Sets state.                                                  #
# Format : setState State                                               #
#=======================================================================#

proc setState {Args} {
   global STATE

   set State [string tolower $Args]

   switch -- $State {
      disabled {set STATE disabled}
      normal   {set STATE normal}
   }
}

#=======================================================================#
# Proc   : setTitle                                                     #
# Purpose: Sets window title bar text.                                  #
# Format : setTitle {Title Options}                                     #
# Options: replace : Replace (rather than append) to window title.      #
#=======================================================================#

proc setTitle {Args} {
   global TITLE

   set Title   [lindex $Args 0]
   set Options [string tolower [lrange $Args 1 end]]

   set Replace 0

   foreach Option $Options {
      switch -- $Option {
         replace {set Replace 1}
      }
   }

   if $Replace {
      wm title . $Title
   } else {
      if {[string compare $Title ""] == 0} {
         wm title . "$TITLE"
      } else {
         wm title . "$TITLE - $Title"
      }
   }
}

#=======================================================================#
# Proc   : validateEntry                                                #
# Purpose: Validates contents of entry.                                 #
# Format : validateEntry Item Mode                                      #
#=======================================================================#

proc validateEntry {Item} {
   global ACTION DATA VALIDATE

   set Name [lindex [split $Item .] 3]

   set ValidationOK 0

   switch -- $VALIDATE($Name) {
      date    {if {[catch {clock scan [formatDate $DATA($Name)]}] == 0} {set ValidationOK 1}}
      default {if {[regexp $VALIDATE($Name) $DATA($Name)]} {set ValidationOK 1}}
   }

   if $ValidationOK {
      setMessage [list "" black]
      return 1
   } else {
      setMessage [list "Invalid $VALIDATE($Name,TEXT)" red]
      return 0
   }
}

#=======================================================================#
# Proc   : login                                                        #
# Purpose: Display and process login screen.                            #
#=======================================================================#

proc login {} {
   addBorder  [list border   10,1 29x7] 
   addEntry   [list User     11,2 9,15 focus]
   addEntry   [list Password 11,5 9,15 secret {login user password}]
   setTitle   {Login}
   setMessage {"Enter User/Password"}
}

#=======================================================================#
# Main code.                                                            #
#=======================================================================#

#-----------------------------------------------------------------------#
# Set parameters.                                                       #
#-----------------------------------------------------------------------#

#set Host [lindex $argv 0]
#set Port [lindex $argv 1]

#-----------------------------------------------------------------------#
# Set platform specific options.                                        #
#-----------------------------------------------------------------------#

set PLATFORM $tcl_platform(platform)

if {[string compare $PLATFORM "unix"] == 0} {
   option add *font {Helvetica 12}
   option add *HighlightThickness 0
}

#-----------------------------------------------------------------------#
# Initialise Globals.                                                   #
#-----------------------------------------------------------------------#

label .dummy

set LOCK             0
set FONT             [.dummy cget -font]
set GROUP            ""
set INITIALISING     1
set TITLE            "Tool"
set VALIDATE(FAILED) 0                          
set XFACTOR          [font measure $FONT "0"]
set YFACTOR          [font metrics $FONT -linespace]
set HEIGHT           [expr {10 * $YFACTOR}]
set WIDTH            [expr {50 * $XFACTOR}]

destroy .dummy

#-----------------------------------------------------------------------#
# Initialise Validations.                                               #
#-----------------------------------------------------------------------#

set VALIDATION(alpha)        {^[a-zA-Z]+$}
set VALIDATION(alphanumeric) {^[a-zA-Z0-9]+$}
set VALIDATION(alphanum)     {^[a-zA-Z0-9]+$}
set VALIDATION(date)         {date}
set VALIDATION(money)        {^[0-9]+[.][0-9][0-9]$}
set VALIDATION(numeric)      {^[0-9]+[.]?[0-9]*$}
set VALIDATION(num)          {^[0-9]+[.]?[0-9]*$}
set VALIDATION(integer)      {^[0-9]+$}
set VALIDATION(int)          {^[0-9]+$}
set VALIDATION(ipaddress)    {^[0-9]{1,3}[.][0-9]{1,3}[.][0-9]{1,3}[.][0-9]{1,3}$}
set VALIDATION(ip)           {^[0-9]{1,3}[.][0-9]{1,3}[.][0-9]{1,3}[.][0-9]{1,3}$}

#-----------------------------------------------------------------------#
# Initialise Date Format.                                               #
#-----------------------------------------------------------------------#

set DATEFORMAT(separator) "/"
set DATEFORMAT(day)       "1"
set DATEFORMAT(month)     "0"
set DATEFORMAT(year)      "2"

#-----------------------------------------------------------------------#
# Initialise Aliases.                                                   #
#-----------------------------------------------------------------------#

set PARSER [interp create -safe]       
                                         
$PARSER alias action          action
$PARSER alias addBorder       addBorder
$PARSER alias addButton       addButton
$PARSER alias addButtons      addButtons
$PARSER alias addCheckButtons addCheckButtons
$PARSER alias addData         addData
$PARSER alias addEntries      addEntries
$PARSER alias addEntry        addEntry
$PARSER alias addLabel        addLabel
$PARSER alias addMenu         addMenu
$PARSER alias addMultiList    addMultiList
$PARSER alias addRadioButtons addRadioButtons
$PARSER alias addValidation   addValidation
$PARSER alias clear           clear
$PARSER alias initGUI         initGUI
$PARSER alias newScreen       newScreen
$PARSER alias remove          remove
$PARSER alias setAction       setAction
$PARSER alias setDateFormat   setDateFormat
$PARSER alias setGroup        setGroup
$PARSER alias setItem         setItem
$PARSER alias setItems        setItems
$PARSER alias setMessage      setMessage
$PARSER alias setSize         setSize
$PARSER alias setState        setState
$PARSER alias setTitle        setTitle
$PARSER alias unlock          unlock
$PARSER alias initMenu        initMenu

#set SERVER [connectToServer $Host $Port]

#tkwait variable INITIALISING

#-----------------------------------------------------------------------#
# Create main display window/frames.                                    #
#-----------------------------------------------------------------------#

set MAIN   .main.main
set RELIEF groove
set STATE  normal

frame .main -width $WIDTH -height $HEIGHT
frame $MAIN -width $WIDTH -height $HEIGHT
frame .message -borderwidth 2 -height 20 -relief raised -width [expr {$WIDTH - 4}]
frame .line -borderwidth 2 -height 2 -relief flat -width $WIDTH
   
label .message.text -text ""
pack  .message.text -side left
pack .line -side top

pack  propagate .main false
pack  propagate $MAIN false
pack  .main

pack  $MAIN    -side top
pack  .message -side bottom -fill x

wm title     . "$TITLE"
wm resizable . 0 0

#-----------------------------------------------------------------------#
# Create "dummy" frame for locking.                                     #
#-----------------------------------------------------------------------#

frame .busylock
bind  .busylock <KeyPress> break
place .busylock -x -2 -y -2

#-----------------------------------------------------------------------#
# Display login screen.                                                 #
#-----------------------------------------------------------------------#

#login

# Uncomment the next line when using "Freewrap"
#show console

#=======================================================================#
# End of script: satguitool.tcl                                         #
#=======================================================================#
